/**************************************************************************
 *
 * lists.c -- List processing functions for boolean queries
 * Copyright (C) 1994  Neil Sharman, Alistair Moffat and Lachlan Andrew
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: lists.cpp 655 1999-10-11 02:58:42Z cs025 $
 *
 **************************************************************************/

#include "sysfuncs.h"

#include "memlib.h"

#include "backend.h"
#include "lists.h"
#include "local_strings.h"
#include "locallib.h"
#include "DocEntry.h"

/*
   $Log$
   Revision 1.1  1999/10/11 02:57:42  cs025
   Base install of MG-PP

   Revision 1.1  1999/08/10 21:18:01  sjboddie
   renamed mg-1.3d directory mg

   Revision 1.2  1999/07/01 09:30:48  rjmcnab
   Changes for better match reporting.

   Revision 1.1  1998/11/17 09:34:47  rjmcnab
   *** empty log message ***

   * Revision 1.3  1994/10/20  03:56:50  tes
   * I have rewritten the boolean query optimiser and abstracted out the
   * components of the boolean query.
   *
   * Revision 1.2  1994/09/20  04:41:37  tes
   * For version 1.1
   *
 */

static char *RCSID = "$Id: lists.cpp 655 1999-10-11 02:58:42Z cs025 $";





/*
 * Make a DocList large enough to hold num entries 
 */
DocList *
DocList_create (int num)
{ int      list_size;
  DocList *d;

  list_size = (num > 0 ? num : 1);
  if (!(d = (DocList *) Xmalloc (sizeof (DocList))))
    return NULL;
  if (!(d->DE = (DocEntry *) Xmalloc (sizeof (DocEntry) * list_size)))
    return (NULL);
  d->num = 0;
  d->total_retrieved = 0;
  d->is_approx = 0;
  d->list_size = list_size;
  return (d);
}

/*
 * Change the size of a DocList so that it can hold num entries 
 */
DocList *
DocList_resize (DocList * d, int num)
{
  if (num > d->list_size)
    { 
      if (d->list_size < 100)
	{
	  d->list_size *= 2;
	}
      else
	{
	  d->list_size += 50;
	}
      
      if (!(d->DE = (DocEntry *) Xrealloc (d->DE, sizeof (DocEntry) * (d->list_size))))
        return (NULL);
    }
  return (d);
}

void DocList_destroy(DocList *d)
{ 
  if (d != NULL)
    {
      if (d->DE != NULL)
	{
	  Xfree(d->DE);
	  Xfree(d);
	}
    }
}


/* IntersectLists () */
/* Returns the intersection of List1 and List2 by deleting from List1 all */
/* entries not in List2.  All of List2 is freed, and the nodes of List1 not */
/* included in the resultant list are freed. */
/* On exit the intersection of the lists is returned, or NULL on empty */
/* intersection. */
/* Assumes that both lists are ordered on increasing Value. */
DocList *
DocList_intersect (DocList * List1, DocList * List2)
{
  DocEntry *s1, *d1, *s2;
  
  if (!List2 || !List1)
    {
      if (List1)
	DocList_destroy(List1);
      if (List2)
	DocList_destroy(List2);
      return (DocList_create (0));
    }
  d1 = s1 = List1->DE;
  s2 = List2->DE;
  while (s1 - List1->DE < List1->num && s2 - List2->DE < List2->num)
    {
      if (s1->DocNum == s2->DocNum)
	*d1++ = *s1++;
      if (s1 - List1->DE < List1->num)
	while (s2 - List2->DE < List2->num && s2->DocNum < s1->DocNum)
	  s2++;
      if (s2 - List2->DE < List2->num)
	while (s1 - List1->DE < List1->num && s1->DocNum < s2->DocNum)
	  s1++;
    }
  List1->num = d1 - List1->DE;
  DocList_destroy(List2);
  return (List1);
}


/* DiffLists () */
/* Returns the difference of List1 and List2 by deleting from List1 all */
/* entries in List2.  All of List2 is freed, and the nodes of List1 not */
/* included in the resultant list are freed. */
/* On exit the differrence of the lists is returned, or NULL on empty */
/* intersection. */
/* Assumes that both lists are ordered on increasing Value. */
DocList *
DocList_diff (DocList * List1, DocList * List2)
{
  DocEntry *s1, *d1, *s2;
  if (!List1)
    {
      if (List2)
	DocList_destroy(List2);
      return (DocList_create (0));
    }
  if (!List2)
    return (List1);
  d1 = s1 = List1->DE;
  s2 = List2->DE;
  while (s1 - List1->DE < List1->num)
    {
      if (s2 - List2->DE == List2->num)
	*d1++ = *s1++;
      else
	{
	  if (s1->DocNum == s2->DocNum)
	    {
	      s1++;
	      s2++;
	    }
	  else if (s1->DocNum < s2->DocNum)
	    *d1++ = *s1++;
	  else
	    s2++;
	}
    }
  List1->num = d1 - List1->DE;
  DocList_destroy(List2);
  return (List1);
}

/* MergeLists () */
/* Finds the union of two lists and returns the head of the new list formed. */
/* Assumes that both lists are ordered on increasing Value. */
DocList *
DocList_merge (DocList * List1, DocList * List2)
{
  DocList *NewList;
  DocEntry *d, *s1, *s2;
  if (!List1)
    {
      if (List2)
	return (List2);
      else
	return (DocList_create (0));
    }
  if (!List2)
    return (List1);
  if (!(NewList = DocList_create (List1->num + List2->num)))
    {
      DocList_destroy(List1);
      DocList_destroy(List2);
      return (NULL);
    }

  s1 = List1->DE;
  s2 = List2->DE;
  d = NewList->DE;
  while (s1 - List1->DE < List1->num || s2 - List2->DE < List2->num)
    {
      if (s1 - List1->DE >= List1->num)
	*d++ = *s2++;
      else if (s2 - List2->DE >= List2->num)
	*d++ = *s1++;
      else if (s1->DocNum == s2->DocNum)
	{
	  *d = *s1++;
	  (d++)->Weight += (s2++)->Weight;
	}
      else if (s1->DocNum < s2->DocNum)
	*d++ = *s1++;
      else
	*d++ = *s2++;
    }
  NewList->num = d - NewList->DE;
  DocList_destroy(List1);
  DocList_destroy(List2);
  return (NewList);
}

/* =========================================================================
 * Function: ReduceList
 * Description: 
 *      Go thru doc-set and get rid of all docs which are unmarked
 *      The motivation for this routine is in the case of CNF queries.
 * Input: 
 * Output: 
 * ========================================================================= */

void
DocList_reduceOrIncluded (DocList * cset)
{
  DocEntry *src;
  DocEntry *dest;

  src = dest = cset->DE;
  for (src = dest = cset->DE; src < cset->DE + cset->num; src++)
    {
      if (src->or_included)
	{
	  src->or_included = 0;
	  *dest++ = *src;
	}
    }
  cset->num = dest - cset->DE;

}				/*ReduceList_OrIncluded */

DocList *
DocList_addDoc (DocList * Docs, int res)
{
  int j;
  /* Search the list for the word */
  for (j = 0; j < Docs->num; j++)
    if (Docs->DE[j].DocNum == res)
      break;

  /* Increment the weight if the word is in the list */
  if (j < Docs->num)
    Docs->DE[j].Weight++;
  else
    {
      /* Create a new entry in the list for the new document */
      Docs = DocList_resize (Docs, Docs->num + 1);
      Docs->DE[Docs->num].DocNum = res;
      Docs->DE[Docs->num].Weight = 1;
      Docs->DE[Docs->num].SeekPos = 0;
      Docs->DE[Docs->num].Len = 0;
      Docs->DE[Docs->num].CompTextBuffer = NULL;
      Docs->DE[Docs->num].Next = NULL;
      Docs->num++;
    }
  return (Docs);
}

void DocList_FreeTextBuffers(DocList *DocList, query_data *qd)
{ int  i;
  long l;

  for (i = 0; i < DocList->size(); i ++)
   {
     l = DocEntry_FreeTextBuffer(&DocList->DE[i]);
     ChangeMemInUse (qd, -l);
   }
}

